/**
    follower.h
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#ifndef __ROBOT_FOLLOWER_H__
#define __ROBOT_FOLLOWER_H__

#include <math.h>
#include "robot.h"
#include "solver.h"
#include "angularsolver.h"
#include <fstream>
#include <iostream>
#include <string>
#include <mbzirc_trackers/TrackerTrajectory.h>
#include <std_msgs/Bool.h>
#include <std_msgs/Int32.h>


namespace mkr { namespace formation {

/**
 * @brief Class Follower represents the main class for control of the robot
 */
class RobotFollower : public Robot {
public:

    /**
     * @brief constructor
     *
     * @param type - type of the robot
     * @param lumInt - the intensity of the light of the robot
     * @param name - the name used for the output file
     * @param position - positions of other robots in the formation
     * @param index - index of the robot
     * @param nRobots - number of the robots in the formation
     * @param vMax - maximum absolute value of the forward velocity
     * @param KMax - maximum absolute value of the curvature
     * @param verMax - maximum absolute value of the ascent velocity
     * @param angMax - maximum absolute value of the angular velocity omega_L
     * @param pitchMax - maximum absolute value of the angular velocity epsilon_L
     * @param lightRatio - ratio between the intensity of the key light and light source of the robot
     */
    RobotFollower(const int type, const double lumInt, const char * name, double *position, const int index, const int nRobots,
                  const double vMax, const double KMax, const double verMax, const double angMax, const double pitchMax, const double lightRatio);

    /**
     * @brief destructor
     */
    virtual ~RobotFollower();

    /**
     * @brief runs the optimization process and the preparations for it
     */
	virtual void run(void);

    /**
     * @brief sets the initial state of the leader
     */
    virtual void setInitialState();

    /**
     * @brief sets the desired trajectory for next N transition points
     * @param leaderCoo - the desired trajectory of the leader
     */
    void setDesiredTrajectory(const double *leaderCoo);

    /**
     * @brief sets the objects of interest
     *
     * @param pointsOfinterest - set of objects of interest
     */
    void setPointsOfInterest(vCoordinates & pointsOfInterest);

    /**
     * @brief computes the deviation of the actual position of the leader from its desired start position
     */
    double distanceFromDesiredPosition();

    /**
     * @brief publishes the desired trajectory of the leader
     */
    void publishDesiredTrajectory(ros::Publisher & pubMessage) const;

    /**
     * @brief publishes the desired angle chi_j between camera optical axis and the xy-plane
     */
    void publishDesiredPitch(ros::Publisher & pubMessage) const;

    /**
     * @brief publishes the boolean which signalize that the robot is ready to start the mission
     */
    void publishReady(ros::Publisher & pubMessage) const;

    /**
     * @brief publishes the desired start position
     */
    void flyToTrajectoryStart(ros::Publisher &pubMessage) const;

    /**
     * @brief ROS publishers
     */
    ros::Publisher pubDesiredTrajectory;
    ros::Publisher pubDesiredPitch;
    ros::Publisher pubReady;

    /**
     * @brief subscribed trajectory of the leader
     */
    double * subscribedTrajectory;
    bool flyNow, start, ready, isTrajectorySet;

private:
    vCoordinates pointsOfInterest;
    ros::NodeHandle nh;
    ros::Subscriber subLeaderActualPosition;
    ros::Subscriber subThisFollowerActualPosition;
    ros::Subscriber subFollower2ActualPosition;
    ros::Subscriber subFollower3ActualPosition;
    ros::Subscriber subDesiredTrajectory;
    ros::Subscriber subFormationStart;
    ros::Subscriber subSyncNumber;
    ros::Subscriber subLeaderPitch;
    ros::Subscriber subFollower2DesiredTrajectory;
    ros::Subscriber subFollower3DesiredTrajectory;
    void leaderPositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void leaderPitchCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg);
    void thisFollowerPositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void secondFollowerPositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void thirdFollowerPositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void secondFollowerTrajectoryCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg);
    void thirdFollowerTrajectoryCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg);
    void desiredTrajectoryCallback(const mbzirc_trackers::TrackerTrajectoryConstPtr &msg);
    void formationStartCallback(const std_msgs::Bool& msg);
    void syncNumberCallback(const std_msgs::Int32& msg);
    void setState(double *x, double stateX, double stateY, double stateZ, double statePhi, double statePsi, double stateEta, int j);
    void computeStates(double *x);
    void reInitialization(void);
    double dist;
    const double type;
    const double lumInt;
    const int indexOfRobot;
    double *actualPositions, *followersTrajectories;
    double sObj, sObjA;
    const double *leaderDesiredTrajectory;
    int nRobots, counter;
    double KMax, vMax, angMax, verMax, pitchMax, lightRatio;
    std::ofstream stream;
    std::ofstream statesOutput;
};

}}

#endif

