/**
    solver.h
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#ifndef SOLVER_H_
#define SOLVER_H_

#include <cstdlib>
#include <iostream>
#include <float.h>
#include "robot.h"
#include "follower.h"
#include "formtime.h"

using namespace std;
using namespace mkr::formation;

/**
 * @brief Class Solver provides all variables and functions essential for the optimization task
 *        for trajectory tracking
 */
class Solver {

public:

    /**
     * @brief constructor
     *
     * @param actual state - actual states of all robots in the formation
     * @param desiredTrajectory - the desired trajectory of the follower
     * @param leaderdesiredTrajectory - the desired trajectory of the leader
     * @param otherFollowersTrajectories - the desired trajectories of others followers
     * @param x - vector of control inputs
     * @param KMax - maximum absolute value of curvature
     * @param vMax - maximum absolute value of forward velocity
     * @param verMax - maximum absolute value of ascent velocity
     * @param actualPositions - actual positions of other robots in the formation
     * @param obstacles - set of all obstacles
     * @param index - the index of the robot
     * @param nRobots - number of robots in the formation
     * @param nObstacles - number of bstacles
     * @param poi - set of objects of interest
     * @param counter - the serial number of currently scanned object of interest
     */
    Solver(const double *actualState, const double *desiredTrajectory, const double *leaderDesiredTrajectory,
           const double *otherFollowersTrajectories, double *x, const double vMax, const double KMax, const double verMax,
           double * actualPositions, double * obstacles, int index, int nRobots, int nObstacles, vCoordinates poi, int counter);

    /**
     * @brief destructor
     */
    ~Solver();

    /**
     * @brief starts the optimization process
     */
    void callSolver(void);

    /**
     * @brief generates the boundary values of optimization vector
     *
     * @param bl - lower boundaries
     * @param bu - upper boundaries
     */
    void generateBoundaryValues(double *bl,double *bu);

    /**
     * @brief computes states of the leader within the optimization process
     *
     * @param x - the vector of control variables
     */
    void computeStates(double *x);

    /**
     * @brief returns the value of the objective function corresponding with the resulting vector of decision variables
     */
    double getObjectiveFunctionValue();

    /**
     * @brief states of the leader computed from the vector of control variables for entire length of control horizon
     */
    double *states;

    /**
     * @brief set of obstacles
     */
    const double *obstacles;

    /**
     * @brief actual positions of other members of the formation
     */
    const double *actualPositions;

    /**
     * @brief desired trajectory of the leader
     */
    const double *leaderDesiredTrajectory;

    /**
     * @brief desired trajectory of other followers
     */
    const double *otherFollowersTrajectories;

    /**
     * @brief index of the robot
     */
    int index;

    /**
     * @brief number of robots
     */
    int nRobots;

    /**
     * @brief number of obstacles
     */
    int nObstacles;

    /**
     * @brief serial number of the currently scanned object of interest
     */
    int counter;

    /**
     * @brief objects of interest
     */
    vCoordinates poi;

private:
	const double *actualState;
	const double *desiredTrajectory;
	double *x;
    const double vMax;
    const double KMax;
    const double angMax;
    const double verMax;
    friend void objF(int nparam,int j, double *x, double *fj, void *cd);
    friend void objF2(int nparam,int j, double *x, double *fj, void *cd);
    friend void constrF1(int nparam, int j, double *x, double *gj, void *cd);
    void setState(double *x, double stateX, double stateY, double stateZ, double statePhi, int j);
    double objf;

};

#endif
 
