/**
    leader.h
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#ifndef __ROBOT_LEADER_H__
#define __ROBOT_LEADER_H__

#include "robot.h"
#include <fstream>
#include <iostream>
#include <std_msgs/Bool.h>

namespace mkr { namespace formation {

/**
 * @brief Class Leader represents the main class for control of the robot
 */
class RobotLeader : public Robot {
public:

    /**
     * @brief constructor
     *
     * @param positions - positions of other robots in the formation
     * @param nRobots - number of the robots in the formation
     * @param vMax - maximum absolute value of the forward velocity
     * @param KMax - maximum absolute value of the curvature
     * @param verMax - maximum absolute value of the ascent velocity
     * @param angMax - maximum absolute value of the angular velocity omega_L
     * @param pitchMax - maximum absolute value of the angular velocity epsilon_L
     */
    RobotLeader(double * positions, const int nRobots, const double vMax, const double KMax, const double verMax,
                const double angMax, const double pitchMax);

    /**
     * @brief destructor
     */
    virtual ~RobotLeader();

    /**
     * @brief runs the optimization process and the preparations for it
     */
	virtual void run(void);

    /**
     * @brief sets the initial state of the leader
     *
     * @param iState - initial state of the robot
     * @param psi - initial psi of the robot
     */
    virtual void setInitialState(Coordinates iState, double psi);

    /**
     * @brief publishes the desired trajectory of the leader
     */
    void publishDesiredTrajectory(ros::Publisher & pubMessage) const;

    /**
     * @brief publishes the boolean which signalize the start of the mission
     */
    void publishStartFollowing(ros::Publisher & pubMessage) const;

    /**
     * @brief publishes the serial number of actual scanned object of interest
     */
    void publishSyncTime(ros::Publisher & pubMessage) const;

    /**
     * @brief publishes the desired angle chi_L between camera optical axis and the xy-plane
     */
    void publishPitchGimbal(ros::Publisher & pubMessage) const;

    /**
     * @brief initializes the vector of control inputs and the the vector of objects of interest at the
     *        beginning of the mission
     */
    void initialization(void);

    /**
     * @brief sets the map
     *
     * @param icoordinates - the desired trajectory
     * @param iCameraPoints - the desired objects of interest
     */
    void setMAP(vCoordinates & icoordinates, vCoordinates & iCameraPoints);

    /**
     * @brief sets the desired trajectory for next N transition points
     */
    void setDesiredTrajectory(void);

    /**
     * @brief publishes the desired start position
     */
    void flyToTrajectoryStart(ros::Publisher & pubMessage) const;

    /**
     * @brief computes the deviation of the actual position of the leader from its desired start position
     */
    double distanceFromDesiredPosition();

    bool ready, finish, readyF1, readyF2, readyF3, start;

    /**
     * @brief ROS publishers
     */
    ros::Publisher pubDesiredTrajectory;
    ros::Publisher pubStartFollowing;
    ros::Publisher pubSyncNumber;
    ros::Publisher pubPosition;
    ros::Publisher pubPitch;

private:
    ros::NodeHandle nh;
    ros::Subscriber subLeaderActualPosition;
    ros::Subscriber subFollower1ActualPosition;
    ros::Subscriber subFollower2ActualPosition;
    ros::Subscriber subFollower3ActualPosition;
    ros::Subscriber subFollower1Ready;
    ros::Subscriber subFollower2Ready;
    ros::Subscriber subFollower3Ready;
    ros::Subscriber subLeaderPitch;
    void leaderPositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void follower1PositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void follower2PositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void follower3PositionCallback(const nav_msgs::OdometryConstPtr &msg);
    void follower1ReadyCallback(const std_msgs::Bool& msg);
    void follower2ReadyCallback(const std_msgs::Bool& msg);
    void follower3ReadyCallback(const std_msgs::Bool& msg);
    void leaderPitchCallback(const nav_msgs::OdometryConstPtr &msg);
    void reInitialization(void);
    void setState(double *x, double stateX, double stateY, double stateZ, double statePhi, double statePsi,
                  double stateEta, int j);
    std::ofstream stream;
    std::ofstream statesOutput;
    std::ofstream desTraj;
    vCoordinates path;
    vCoordinates pointsOfInterest;
    double KMax, vMax, angMax, verMax, pitchMax;
    double sObj, sObjA;
    int syncNumber, nRobots;
};

}}

#endif

