/**
    leadersolver.h

    @author Vit Kratky
    @version 1.3 20/05/17
*/
#ifndef LEADERSOLVER_H_
#define LEADERSOLVER_H_
#include "leader.h"

using namespace std;
using namespace mkr::formation;

/**
 * @brief Class LeaderSolver provides all variables and functions essential for the optimization task
 *        for trajectory tracking
 */
class LeaderSolver {
private:
	const double *desiredTrajectory;
	double *x;
    const double KMax, vMax, verMax;
    friend void obj(int nparam,int j, double *x, double *fj, void *cd);
    friend void constr(int nparam, int j, double *x, double *gj, void *cd);
    void setState(double *x, double stateX, double stateY, double stateZ, double statePhi, int j);
    double objf;
public:

    /**
     * @brief constructor
     *
     * @param actual state - actual states of all robots in the formation
     * @param desiredTrajectory - the desired trajectory of the leader
     * @param x - vector of control inputs
     * @param KMax - maximum absolute value of curvature
     * @param vMax - maximum absolute value of forward velocity
     * @param verMax - maximum absolute value of ascent velocity
     * @param obstacles - set of all obstacles
     * @param nRobots - number of robots in the formation
     * @param nObstacles - number of bstacles
     * @param poi - set of objects of interest
     */
    LeaderSolver(const double *actualState, const double *desiredTrajectory,
           double *x, const double KMax, const double vMax, const double verMax,
           double * obstacles, int nRobots, int nObstacles, vCoordinates poi);

    /**
     * @brief destructor
     */
    ~LeaderSolver();

    /**
     * @brief starts the optimization process
     */
    void callSolver(void);

    /**
     * @brief generates the boundary values of optimization vector
     *
     * @param bl - lower boundaries
     * @param bu - upper boundaries
     */
    void generateBoundaryValues(double *bl,double *bu);

    /**
     * @brief computes states of the leader within the optimization process
     *
     * @param x - the vector of control variables
     */
    void computeStates(double *x);

    /**
     * @brief returns the value of the objective function corresponding with the resulting vector of decision variables
     */
    double getObjectiveFunctionValue();

    /**
     * @brief states of the leader computed from the vector of control variables for entire length of control horizon
     */
    double *states;

    /**
     * @brief actual states of all robots in the formation
     */
    const double *actualState;

    /**
     * @brief set of obstacles
     */
    const double *obstacles;

    /**
     * @brief objects of interest
     */
    vCoordinates poi;

    int nRobots;
    int nObstacles;

};

#endif
 
