/**
    leader.cc
    author of the original version: Vojtech Spurny
    @author Vit Kratky
    @version 1.3 20/05/17
*/
#include "leader.h"
#include "leadersolver.h"
#include "angularsolver.h"
#include "map.h"
#include <ros/ros.h>
#include <sstream>
#include <string.h>
#include <nav_msgs/Odometry.h>
#include <visualization_msgs/Marker.h>
#include <tf/transform_datatypes.h>
#include <mbzirc_trackers/TrackerTrajectory.h>
#include "std_msgs/Bool.h"
#include "std_msgs/Int32.h"
#include "std_msgs/Float32.h"

using namespace std;
using namespace mkr::formation;

RobotLeader::RobotLeader(double* positions, const int nRobots, const double vMax, const double KMax, const double verMax,
                         const double angMax, const double pitchMax) : Robot() {
    ready=false;
    nh = ros::NodeHandle("~");
    //stream.open("/home/vitkratky/cvut/project/data/leader_time.txt");
    //statesOutput.open("/home/vitkratky/cvut/project/data/leader_states.txt");
    //desTraj.open("/home/vitkratky/cvut/project/data/leader_trajectory.txt");
    subLeaderActualPosition = nh.subscribe("actual_position", 1, &RobotLeader::leaderPositionCallback, this);
    subFollower1ActualPosition = nh.subscribe("follower1_position", 1, &RobotLeader::follower1PositionCallback, this);
    subFollower2ActualPosition = nh.subscribe("follower2_position", 1, &RobotLeader::follower2PositionCallback, this);
    subFollower3ActualPosition = nh.subscribe("follower3_position", 1, &RobotLeader::follower3PositionCallback, this);
    pubDesiredTrajectory = nh.advertise<mbzirc_trackers::TrackerTrajectory>("desired_trajectory", 1);
    pubStartFollowing = nh.advertise<std_msgs::Bool>("start_following", 1);
    pubSyncNumber = nh.advertise<std_msgs::Int32>("sync_time", 1);
    pubPosition = nh.advertise<mbzirc_trackers::TrackerTrajectory>("leader_position", 1);
    pubPitch = nh.advertise<mbzirc_trackers::TrackerTrajectory>("leader_pitch", 1);
    subFollower3Ready = nh.subscribe("follower3_ready", 1, &RobotLeader::follower3ReadyCallback, this);
    subFollower2Ready = nh.subscribe("follower2_ready", 1, &RobotLeader::follower2ReadyCallback, this);
    subFollower1Ready = nh.subscribe("follower1_ready", 1, &RobotLeader::follower1ReadyCallback, this);
    this->nRobots = nRobots;
    this->KMax = KMax;
    this->angMax = angMax;
    this->vMax = vMax;
    this->verMax = verMax;
    this->pitchMax = pitchMax;
    actualState = &positions[0];
    finish = false;
    start = false;
    readyF1 = false;
    readyF2 = false;
    readyF3 = false;
    x = (double*)malloc(LENGTH_HORIZONT*(N_PARAMS)*sizeof(double));
    syncNumber = 0;
}

RobotLeader::~RobotLeader() {
    //stream.close();
    //statesOutput.close();
    //desTraj.close();
    free(actualState);
    free(x);
    free(desiredTrajectory);
    if(obstacles != NULL){
        free(obstacles);
    }
}

void RobotLeader::leaderPositionCallback(const nav_msgs::OdometryConstPtr &msg){
    if(!start){
    actualState[0] = msg->pose.pose.position.x;
    actualState[1] = msg->pose.pose.position.y;
    actualState[7] = msg->pose.pose.position.z;
    cout << actualState[0] << " " << actualState[1] << " " << actualState[3] << endl;
    }
}

void RobotLeader::follower1ReadyCallback(const std_msgs::Bool& msg){
    readyF1 = msg.data;
}

void RobotLeader::follower2ReadyCallback(const std_msgs::Bool& msg){
    readyF2 = msg.data;
}

void RobotLeader::follower3ReadyCallback(const std_msgs::Bool& msg){
    readyF3 = msg.data;
}

void RobotLeader::leaderPitchCallback(const nav_msgs::OdometryConstPtr &msg){
    if(!start){
    actualState[9] = tf::getYaw(msg->pose.pose.orientation);//unused
    }
}

void RobotLeader::publishDesiredTrajectory(ros::Publisher & pubMessage) const {
    mbzirc_trackers::TrackerTrajectory new_traj;
    mbzirc_trackers::TrackerPoint point;
    for(int i = 0; i < LENGTH_HORIZONT; i++){
        point.x = desiredTrajectory[i];
        point.y = desiredTrajectory[LENGTH_HORIZONT+i];
        point.z = desiredTrajectory[4*LENGTH_HORIZONT+i];
        point.yaw = desiredTrajectory[3*LENGTH_HORIZONT+i];
        new_traj.points.push_back(point);
    }
    new_traj.header.stamp = ros::Time::now();
    new_traj.header.frame_id = "local_origin";
    new_traj.use_yaw = true;
    new_traj.fly_now = true;//can be always true
    pubMessage.publish(new_traj);
}

void RobotLeader::publishPitchGimbal(ros::Publisher & pubMessage) const {
    mbzirc_trackers::TrackerTrajectory new_traj;
    mbzirc_trackers::TrackerPoint point;
    for(int i = 0; i < LENGTH_HORIZONT; i++){
        point.yaw = desiredTrajectory[5*LENGTH_HORIZONT+i];
        new_traj.points.push_back(point);
    }
    new_traj.header.stamp = ros::Time::now();
    new_traj.header.frame_id = "local_origin";
    new_traj.use_yaw = true;
    new_traj.fly_now = true;
    pubMessage.publish(new_traj);
}

void RobotLeader::publishStartFollowing(ros::Publisher & pubMessage) const{
    std_msgs::Bool msg;
    msg.data = start;
    pubMessage.publish(msg);
}

void RobotLeader::publishSyncTime(ros::Publisher &pubMessage) const{
    std_msgs::Int32 msg;
    msg.data = syncNumber;
    pubMessage.publish(msg);
}

void RobotLeader::flyToTrajectoryStart(ros::Publisher &pubMessage) const {
    mbzirc_trackers::TrackerTrajectory new_traj;
    mbzirc_trackers::TrackerPoint point;
    point.x = desiredTrajectory[0];
    point.y = desiredTrajectory[LENGTH_HORIZONT];
    point.z = desiredTrajectory[4*LENGTH_HORIZONT];
    point.yaw = desiredTrajectory[3*LENGTH_HORIZONT];
    for (int i = 0; i < LENGTH_HORIZONT; ++i) {
        new_traj.points.push_back(point);
    }
    new_traj.header.stamp = ros::Time::now();
    new_traj.header.frame_id = "local_origin";
    new_traj.use_yaw = true;//should be always true
    new_traj.fly_now = true;
    pubMessage.publish(new_traj);
    cout << "Fly to trajectory start" << endl;
}

void RobotLeader::follower1PositionCallback(const nav_msgs::OdometryConstPtr &msg){
    actualState [N_STATES+0] = msg->pose.pose.position.x;
    actualState[N_STATES+1] = msg->pose.pose.position.y;
    actualState[N_STATES+7] = msg->pose.pose.position.z;
    actualState[N_STATES+3] = fmod(tf::getYaw(msg->pose.pose.orientation)+2*PI, 2*PI);
}

void RobotLeader::follower2PositionCallback(const nav_msgs::OdometryConstPtr &msg){
    actualState [2*N_STATES+0] = msg->pose.pose.position.x;
    actualState[2*N_STATES+1] = msg->pose.pose.position.y;
    actualState[2*N_STATES+7] = msg->pose.pose.position.z;
    actualState[2*N_STATES+3] = fmod(tf::getYaw(msg->pose.pose.orientation)+2*PI, 2*PI);
}

void RobotLeader::follower3PositionCallback(const nav_msgs::OdometryConstPtr &msg){
    actualState [3*N_STATES+0] = msg->pose.pose.position.x;
    actualState[3*N_STATES+1] = msg->pose.pose.position.y;
    actualState[3*N_STATES+7] = msg->pose.pose.position.z;
    actualState[3*N_STATES+3] = fmod(tf::getYaw(msg->pose.pose.orientation)+2*PI, 2*PI);
}



int main(int argc, char ** argv) {
    int nRobots;
    double vMax, KMax, verMax, angMax, pitchMax;
    ros::init(argc, argv, "leader_camera");
    ROS_INFO("The node started!");
    string addressMap, addressObstacles,stringName;
    ros::NodeHandle nh("~");
    ros::NodeHandle privateNodeHandle("~");
    privateNodeHandle.param("name", stringName, string("{0,1,2}"));
    privateNodeHandle.param("addressMap", addressMap, string("cfg/LoadMap"));
    privateNodeHandle.param("addressObstacles", addressObstacles, string("cfg/LoadObstacles"));
    privateNodeHandle.param("number_of_robots", nRobots, int(3));
    privateNodeHandle.param("v_max", vMax, double(0.5));
    privateNodeHandle.param("ver_max", verMax, double(0.5));
    privateNodeHandle.param("k_max", KMax, double(2));
    privateNodeHandle.param("ang_max", angMax, double(0.1));
    privateNodeHandle.param("pitch_max", pitchMax, double(0.1));
    Map map;
    map.loadMap(addressMap.c_str());
    map.loadObstacles(addressObstacles.c_str());
    //map.segmentationMap(TS*VELOCITY_LEADER);
    double * positionsOfRobots;
    positionsOfRobots= (double*)calloc ((nRobots)*N_STATES,sizeof(double));
    RobotLeader * leader = new RobotLeader(positionsOfRobots, nRobots, KMax, vMax, verMax, angMax, pitchMax);
    leader->setMAP(map.getCoordinates(), map.getPointsOfInterest());
    leader->setObstacles(map.getObstacles());
    leader->setInitialState(map.getCoordinates()[0], map.getPointsOfInterest()[0].phi);//it has to move
    leader->initialization();
    leader->setDesiredTrajectory();
    ros::Rate loop_rate(1/(TS*N)); // Frequency in Hz /****
    leader->resetTime();
    while (nh.ok() && !leader->start) {
        ros::spinOnce();
        if(leader->distanceFromDesiredPosition() < 1){
            leader->ready = true;
        }
        if(nRobots == 4 && leader->ready && leader->readyF1 && leader->readyF2 && leader->readyF3){
            leader->start = true;
        }else if(nRobots == 3 && leader->ready && leader->readyF1 && leader->readyF2){
            leader->start = true;
        }else if(nRobots == 2 && leader->ready && leader->readyF1){
            leader->start = true;
        }else if(nRobots == 1 && leader->ready){
            leader->start = true;
        }else{
            ROS_INFO("Waiting for followers.");
        }
        leader->flyToTrajectoryStart(leader->pubDesiredTrajectory);
        leader->publishStartFollowing(leader->pubStartFollowing);
        leader->publishSyncTime(leader->pubSyncNumber);
        loop_rate.sleep();
    }
    ros::spinOnce();
    leader->run();
    while (nh.ok() && !leader->finish) {
        leader->publishDesiredTrajectory(leader->pubDesiredTrajectory);
        leader->publishSyncTime(leader->pubSyncNumber);
        ros::spinOnce();
        leader->run();
        loop_rate.sleep();
    }
    ros::spinOnce();
    ros::Duration(10);
    free(positionsOfRobots);
    return EXIT_SUCCESS;
}

double RobotLeader::distanceFromDesiredPosition(){
    return sqrt(pow(actualState[0] - desiredTrajectory[0], 2)+pow(actualState[1] -
            desiredTrajectory[LENGTH_HORIZONT], 2)+pow(actualState[7] -
            desiredTrajectory[4*LENGTH_HORIZONT], 2)) + fmod(desiredTrajectory[3*LENGTH_HORIZONT]-actualState[3] +4*PI, 2*PI);
}

void RobotLeader::initialization(void){//
	for (int j = 0; j < LENGTH_HORIZONT; j++) {
      x[j] = VELOCITY_LEADER;
	  x[LENGTH_HORIZONT+j] = 0;
      x[2*LENGTH_HORIZONT+j] = 0;
      x[3*LENGTH_HORIZONT+j] = 0;
      x[4*LENGTH_HORIZONT+j] = 0;
	}
	vCoordinates newPath;
    vCoordinates newPoints;
	for (unsigned int j = 0; j < path.size()-1; j++) {
	    newPath.push_back(path[j+1]);
        newPoints.push_back(pointsOfInterest[j+1]);
	}
    pointsOfInterest = newPoints;
	path = newPath;
}

void RobotLeader::reInitialization(void){//
    for (int j = 0; j < LENGTH_HORIZONT-N; j++) {
        x[j] = x[j+N];
        x[LENGTH_HORIZONT+j] = x[LENGTH_HORIZONT+j+N];
        x[2*LENGTH_HORIZONT+j] = x[2*LENGTH_HORIZONT+j+N];
        x[3*LENGTH_HORIZONT+j] = x[3*LENGTH_HORIZONT+j+N];
        x[4*LENGTH_HORIZONT+j] = x[4*LENGTH_HORIZONT+j+N];
    }
    vCoordinates newPath;
    vCoordinates newPoints;
    for (unsigned int j = 0; j < path.size()-N; j++) {
        newPath.push_back(path[j+N]);
        newPoints.push_back(pointsOfInterest[j+N]);
    }
	path = newPath;
    pointsOfInterest = newPoints;
	for (int j = LENGTH_HORIZONT-N; j < LENGTH_HORIZONT; j++) {
        x[j] = VELOCITY_LEADER;
        x[LENGTH_HORIZONT+j] = 0;
        x[2*LENGTH_HORIZONT+j] = 0;
        x[3*LENGTH_HORIZONT+j] = 0;
        x[4*LENGTH_HORIZONT+j] = 0;
    }
	setDesiredTrajectory();
}

void RobotLeader::setDesiredTrajectory(void){
    double xy_dist;
    for (int j = 0; j < LENGTH_HORIZONT; j++) {
        xy_dist = sqrt(pow(pointsOfInterest[j].x-path[j].x, 2)+pow(pointsOfInterest[j].y-path[j].y, 2));
        desiredTrajectory[j] = path[j].x;
        desiredTrajectory[LENGTH_HORIZONT+j] = path[j].y;
        desiredTrajectory[3*LENGTH_HORIZONT+j] = pointsOfInterest[j].phi;
        desiredTrajectory[4*LENGTH_HORIZONT+j] = path[j].z;
        desiredTrajectory[5*LENGTH_HORIZONT+j] = atan2(path[j].z - pointsOfInterest[j].z,
                                                xy_dist);
    }
}

void RobotLeader::setMAP(vCoordinates & icoordinates, vCoordinates & iCameraPoints) {
	path = icoordinates;
    pointsOfInterest = iCameraPoints;
}

void RobotLeader::setInitialState(Coordinates iState, double psi) {//hopefully unused
	actualState[0] = iState.x;
    actualState[1] = iState.y;
	actualState[2] = iState.phi;
    actualState[3] = psi;
    actualState[7] = iState.z;
    actualState[9] = iState.eta;
}

void  RobotLeader::setState(double *x, double stateX, double stateY, double stateZ, double statePhi, double statePsi, double stateEta, int j){
    if (fabs(x[LENGTH_HORIZONT+j]) < (1E-6F)) { //if K==0
       desiredTrajectory[j]= stateX + x[j]*cos(statePhi)*TS;
       desiredTrajectory[LENGTH_HORIZONT+j]= stateY + x[j]*sin(statePhi)*TS;
	} else {
       desiredTrajectory[j]= stateX + (-sin(statePhi)+sin(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
       desiredTrajectory[LENGTH_HORIZONT+j]= stateY - (-cos(statePhi)+cos(statePhi+x[LENGTH_HORIZONT+j]*x[j]*TS)) / x[LENGTH_HORIZONT+j];
	}
    desiredTrajectory[2*LENGTH_HORIZONT+j] = statePhi + x[LENGTH_HORIZONT+j]*x[j]*TS;
    desiredTrajectory[3*LENGTH_HORIZONT+j] = fmod(statePsi + x[3*LENGTH_HORIZONT+j]*TS + 2*PI, 2*PI);
    desiredTrajectory[4*LENGTH_HORIZONT+j] = stateZ + x[2*LENGTH_HORIZONT+j]*TS;
    desiredTrajectory[5*LENGTH_HORIZONT+j] = stateEta + x[4*LENGTH_HORIZONT+j]*TS;
}

void RobotLeader::run() {
	 if (path.size() - N >= LENGTH_HORIZONT){ 
       reInitialization();
       cout << "poi: x = " << pointsOfInterest[0].x << ", y= " << pointsOfInterest[0].y << endl;
       runtime.init();
       LeaderSolver * solver = new LeaderSolver (actualState, desiredTrajectory, x, KMax, vMax, verMax,
                                                 obstacles, nRobots, nObstacle, pointsOfInterest);
       AngularSolver * angSolver = new AngularSolver(actualState, x, angMax, pitchMax, pointsOfInterest);
       runtime.now();
       sObj = solver->getObjectiveFunctionValue();
       sObjA = angSolver->getObjectiveFunctionValue();
       //stream << runtime.sec << " " << runtime.nsec << " " << sObj << " " << sObjA << endl;
	   for (int j = 0; j < LENGTH_HORIZONT; j++) {
          if (j == 0) {
              setState(x, actualState[0], actualState[1], actualState[7], actualState[2], actualState[3],
                      actualState[9], j);
          }else {
              setState(x, desiredTrajectory[j-1], desiredTrajectory[LENGTH_HORIZONT+j-1],
                      desiredTrajectory[4*LENGTH_HORIZONT+j-1], desiredTrajectory[2*LENGTH_HORIZONT+j-1],
                      desiredTrajectory[3*LENGTH_HORIZONT+j-1], desiredTrajectory[5*LENGTH_HORIZONT+j-1], j);
          }
       }
	   nIter = 0;
	   delete solver;
       delete angSolver;
     } else {
        input.forwardVelocity = 0.0;
	    input.curvature = 0.0;
        input.angularVelocity = 0.0;
        ROS_INFO("FINAL POSITION");
        finish=true;
        start = false;
        publishStartFollowing(pubStartFollowing);
	  }
	  
    for (int nIter = 0; nIter < N; nIter++) {
        input.forwardVelocity = x[nIter];
        input.angularVelocity = x[3*LENGTH_HORIZONT+nIter];
        input.curvature = x[LENGTH_HORIZONT+nIter];
        input.ascentVelocity = x[2*LENGTH_HORIZONT+nIter];
        input.pitchVelocity = x[4*LENGTH_HORIZONT+nIter];
        cout << input.forwardVelocity << " " << input.curvature << " " << input.angularVelocity << " " << input.ascentVelocity << " " << input.pitchVelocity << endl;
        move();
        //desTraj << actualState[0] << " " << actualState[1] << " " << actualState[7] << " " << actualState[3] << endl;
        /*statesOutput << actualState[0] << " " << actualState[1] << " " << " " << actualState[7] << " " << actualState[2] << " " <<
                     actualState[3] << " " << actualState[4] << " " << actualState[5] << " " <<
                     actualState[6] << " " << actualState[8] << " " << actualState[9] << " " << actualState[10] << endl;*/
        syncNumber++;
        cout << syncNumber << endl;
    }
}

